// Library for nRF24L01+ Receiver.
// 2016-11-12  T. Nakagawa

#include <SPI.h>
#include "NRF24L01.h"

uint8_t NRF24L01::readStatus() {
  digitalWrite(cs_pin_, LOW);
  const uint8_t result = SPI.transfer(0xff);
  digitalWrite(cs_pin_, HIGH);
  return result;
}

uint8_t NRF24L01::readRegister(uint8_t reg) {
  digitalWrite(cs_pin_, LOW);
  SPI.transfer(reg);
  const uint8_t result = SPI.transfer(0xff);
  digitalWrite(cs_pin_, HIGH);
  return result;
}

void NRF24L01::readRegister(uint8_t reg, uint8_t *data, uint8_t len) {
  digitalWrite(cs_pin_, LOW);
  SPI.transfer(reg);
  do {
    *data++ = SPI.transfer(0xff);
  } while (--len);
  digitalWrite(cs_pin_, HIGH);
}

void NRF24L01::writeRegister(uint8_t reg, uint8_t val) {
  digitalWrite(cs_pin_, LOW);
  SPI.transfer(0x20 | reg);
  SPI.transfer(val);
  digitalWrite(cs_pin_, HIGH);
}

void NRF24L01::writeRegister(uint8_t reg, const uint8_t *data, uint8_t len) {
  digitalWrite(cs_pin_, LOW);
  SPI.transfer(0x20 | reg);
  do {
    SPI.transfer(*data++);
  } while (--len);
  digitalWrite(cs_pin_, HIGH);
}

NRF24L01::NRF24L01(uint8_t ce_pin, uint8_t cs_pin) : ce_pin_(ce_pin), cs_pin_(cs_pin) {
  pinMode(ce_pin_, OUTPUT);
  pinMode(cs_pin_, OUTPUT);
  digitalWrite(ce_pin_, LOW);
  digitalWrite(cs_pin_, HIGH);
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(SPI_MODE0);
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  
  delayMicroseconds(100000UL);  // Power on reset.
  writeRegister(0x00, 0x00);  // PWR_UP=0.
  writeRegister(0x00, 0x0b);  // PWR_UP=1 and PRIM_RX=1.
  delayMicroseconds(5000UL);  // Tpd2stby.
  writeRegister(0x01, 0x00);  // Disable auto acknowledgment.
  writeRegister(0x04, 0x00);  // Disable auto retransmit.
  writeRegister(0x1c, 0x00);  // Disable dynamic payload length.
  writeRegister(0x1d, 0x00);  // Disable dynamic payload length and Payload with ACK.
}

void NRF24L01::enable() {
  digitalWrite(ce_pin_, HIGH);  // CE=1.
  delayMicroseconds(130UL);  // RX Settling.
}

void NRF24L01::disable() {
  digitalWrite(ce_pin_, LOW);  // CE=0.
}

boolean NRF24L01::empty() {
  return (boolean)(readRegister(0x17) & 0x01);
}

void NRF24L01::setCRC(CRC crc) {
  writeRegister(0x00, crc);
}

void NRF24L01::setDataRate(DataRate dr) {
  writeRegister(0x06, dr);
}

void NRF24L01::setChannel(uint8_t ch) {
  writeRegister(0x05, ch);
}

void NRF24L01::setPipes(uint8_t pipes) {
  writeRegister(0x02, (uint8_t)(1 << pipes) - 1);  // Enable data pipe 0~(pipes - 1).
}

void NRF24L01::setAddressWidth(uint8_t aw) {
  writeRegister(0x03, aw - 2);
}

void NRF24L01::setAddress(uint8_t pipe, const uint8_t *adrs) {
  if (pipe < 2) {
    writeRegister(0x0a + pipe, adrs, readRegister(0x03) + 2);
  } else {
    writeRegister(0x0a + pipe, adrs, 1);
  }
}

void NRF24L01::setPayloadSize(uint8_t size) {
  for (uint8_t i = 0; i < 6; i++) {
    writeRegister(0x11 + i, size);
  }
}

uint8_t NRF24L01::pipeNumber() {
  return ((readStatus() >> 1) & 0x07);
}

void NRF24L01::read(uint8_t *data) {
  readRegister(0x61, data, readRegister(0x11));  // readRegister() works also for R_RX_PAYLOAD.
  writeRegister(0x07, 0xff);  // Clear IRQ bits.
}

void NRF24L01::flush() {
  writeRegister(0xe2, 0xff);  // writeRegister() works also for FLUSH_RX.
}

boolean NRF24L01::testRPD() {
  return (boolean)(readRegister(0x09) & 0x01);
}

#ifdef DEBUG
void NRF24L01::dumpRegisters() {
  for (uint8_t i = 0x00; i <= 0x1d; i++) {
    if (0x18 <= i && i <= 0x1b) {
      continue;
    }
    Serial.print(i, HEX);
    Serial.print(": ");
    if (i == 0x0a || i == 0x0b || i == 0x10) {
      uint8_t data[5];
      readRegister(i, data, 5);
      for (uint8_t j = 0; j < 5; j++) {
        Serial.print(data[j], HEX);
        Serial.print(" ");
      }
      Serial.println("");
    } else {
      Serial.println(readRegister(i), HEX);
    }
  }
}
#endif
